import os
import sys

import click

from ._common import require_app_module, APP_PATH
from .run import Run
from .shell import Shell
from .test import Test


def main():
    sys.path.insert(0, os.getcwd())

    os.environ.setdefault('MEST_CLI', 'true')
    obj = {}
    try:
        main_app = __main_app()
        obj['app'] = main_app
    except Exception as e:
        obj['err'] = e

    mest_cli(obj=obj)


@click.group()
def mest_cli():
    pass


def __main_app():
    import importlib.util

    if os.path.isfile(APP_PATH):
        spec = importlib.util.spec_from_file_location('mest_app', APP_PATH)
        app_main = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(app_main)
        return app_main


Run.register(mest_cli)
Test.register(mest_cli)
Shell.register(mest_cli)



from flask import request


def get_request_info():
    if not bool(request):
        return None

    result = dict(
        method=request.method,
        url=request.url,
        headers=dict(request.headers),
        url_params=request.args,
        form_params=request.form,
        cookies=request.cookies
    )

    if request.data:
        result['payload'] = request.data.decode('utf-8')

    return result



import datetime
import logging
import traceback


def _run_enrichment_method(enrichment_method):
    try:
        return enrichment_method()
    except Exception as e:
        return {'error': str(e)}


class DictFormatter(logging.Formatter):
    def __init__(self, app_name, enrichment_methods=None, **kwargs):
        super().__init__(**kwargs)

        self.app_name = app_name
        self.enrichment_methods = enrichment_methods

    def format(self, record):
        message_json = dict()

        self.add_exception_info(record, message_json)

        message_json['time'] = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')

        message_json['args'] = dict()
        self.add_custom_message(record, message_json)
        self.add_enrichment(message_json)

        message_json['logger'] = self.app_name
        message_json['level'] = record.levelname
        message_json['log_trace'] = record.pathname
        if not any(message_json['args']):
            message_json.pop('args')

        return message_json

    def add_enrichment(self, message_json):
        if not self.enrichment_methods:
            return

        for field, enrichment_method in self.enrichment_methods.items():
            enrichment_value = _run_enrichment_method(enrichment_method)

            if enrichment_value is not None:
                message_json['args'][field] = enrichment_value

    @staticmethod
    def add_exception_info(record, message_json):
        if bool(record.exc_info) and any(record.exc_info):
            message_json['error_class_name'] = record.exc_info[1].__class__.__name__
            message_json['error_message'] = str(record.exc_info[1])
            message_json['backtrace'] = '\n'.join(traceback.format_exception(*record.exc_info))

    @staticmethod
    def add_custom_message(record, message_json):
        if isinstance(record.msg, dict):
            message_json['msg'] = '{} {}'.format(record.msg.get('message', ''),
                                                 record.msg.get('msg', '')).strip()

            # remove messages from args
            args = {k: record.msg[k] for k in record.msg if k not in ['message', 'msg']}
            message_json['args'].update(args)
        else:
            message_json['msg'] = str(record.msg)


from functools import wraps


def singleton():
    def decorator(cls):
        cls.instance = None

        @wraps(cls)
        def wrapper():
            if cls.instance is None:
                cls.instance = cls()

            return cls.instance

        wrapper.original_class = cls
        return wrapper

    return decorator



import unittest

from mest.utils import singleton

PROP_VALUE = 10


class MestCoreUtilsTest(unittest.TestCase):
    """Test cases for mest utils"""

    def setUp(self):
        pass

    def test_singleton(self):
        obj1 = MyClass()
        new_value = PROP_VALUE * 2
        obj1.prop = new_value

        obj2 = MyClass()

        self.assertEqual(obj1, obj2)
        self.assertEqual(obj2.prop, new_value)

    def test_2_singletons(self):
        a = MyClass()
        b = MyOtherClass()

        self.assertNotEqual(a, b)
        self.assertNotEqual(type(a), type(b))


@singleton()
class MyClass(object):
    def __init__(self):
        self.prop = 0


@singleton()
class MyOtherClass(object):
    def __init__(self):
        self.x = 0
        self.y = 0



import datetime
import json

from flask import Blueprint, current_app

from mest.app._state import runtime_state
from ._request_parser import validate_params


class Router(object):
    def __init__(self, name):
        self._blueprint = Blueprint(name, __name__)

    def route(self, url, method='POST', **options):
        def decorator(func):
            methods = options.pop('methods', [method])

            blueprint_decorator = self._blueprint.route(url, methods=methods, endpoint=func.__name__, **options)

            return blueprint_decorator(func)

        return decorator

    def add_predict_route(self, model_instance, schema, url='/predict', method='POST'):
        @self.route(url, method=method)
        @validate_params(schema=schema)
        def predict(**params):
            try:
                result = model_instance.predict(**params)

                return json.dumps(result)
            except Exception as e:
                response = {'message': str(e)}

                return json.dumps(response), 500

    def add_ping_route(self, url='/ping', method='GET'):
        @self.route(url=url, method=method)
        def ping():
            msg = 'Pong from {}! - {}'.format(current_app.config['MEST'].service_name,
                                              datetime.datetime.now())

            status_code = 503 if runtime_state.is_shutting_down() else 200

            return json.dumps({'message': msg}), status_code



import abc


class GenericModel(object):
    def __init__(self):
        self.model = None

    @abc.abstractmethod
    def init(self, local_model_dir_path):
        pass

    def predict(self, **kwargs):
        inference_payload = self.pre_process(**kwargs)

        prediction = self.__invoke_with_expansion(method=self.infer,
                                                  payload=inference_payload)

        return self.__invoke_with_expansion(method=self.post_process,
                                            payload=prediction)

    def pre_process(self, **kwargs):
        return kwargs

    @abc.abstractmethod
    def infer(self, *args, **kwargs):
        pass

    def post_process(self, *args, **kwargs):
        if args:
            if len(args) == 1:
                return args[0]
            else:
                return args
        return kwargs

    def validate(self):
        pass

    @property
    def is_loaded(self):
        return self.model is not None

    @staticmethod
    def __invoke_with_expansion(method, payload):
        payload_type = type(payload)

        if payload_type == dict:
            return method(**payload)
        elif payload_type == list:
            return method(*payload)
        else:
            return method(payload)


class ModelValidationError(RuntimeError):
    pass



import os

from mest.models import GenericModel
from mest.app import MestConfig


class MyTestModel(GenericModel):
    def init(self, path):
        file_path = os.path.join(path, 'model.txt')
        with open(file_path, 'r') as fs:
            self.model = fs.read()

    def infer(self, *args, **kwargs):
        return self.model


def current_file_path():
    return os.path.dirname(os.path.abspath(__file__))


def generate_mest_config() -> MestConfig:
    return MestConfig(service_name='test_mest',
                      listen_port=1234)


def generate_mest_config_with_model() -> MestConfig:
    conf = generate_mest_config()
    conf.local_model_directory_path = os.path.join(current_file_path(), '_models')

    return conf



import os
import sys

import click

from mest.app.mest_conf import Environment
from ._common import require_mest_app, APP_PATH


class Run(object):
    @staticmethod
    def register(mest_cli):
        @mest_cli.group(name='run', help='Run development/gunicorn server')
        def run():
            os.environ['MEST_CLI'] = 'false'

        @run.command(name='dev', help='Run development server')
        @click.option('--debug/--no-debug', default=False)
        @click.pass_context
        def dev(ctx, debug):
            if debug:
                os.environ.setdefault('FLASK_ENV', Environment.DEVELOPMENT)
            os.environ.setdefault('PRETTY_LOG', 'true')
            mest = require_mest_app(ctx)
            mest.setup()
            mest.run()

        @run.command(name='gunicorn', help='Run gunicorn server')
        @click.pass_context
        def gunicorn(ctx):
            mest = require_mest_app(ctx)
            listen_port = int(mest.config.listen_port)

            cmd = 'gunicorn -w 1 -b 0.0.0.0:{listen_port} {app_module}:wsgi --graceful-timeout 30'.format(
                listen_port=listen_port,
                app_module=__parse_app_module())
            exit_code = os.system(cmd)
            sys.exit(exit_code)

        def __parse_app_module():
            return APP_PATH.split('.')[0]



import time
from multiprocessing.dummy import Pool as ThreadPool

from flask import g, request


class RequestStatter(object):
    def __init__(self, app=None, service_name='py_model', config=None, report_method=None):
        self.config = config
        self.prefix = 'services.{}'.format(service_name)
        self.pool = ThreadPool(1)
        self.report_method = report_method

        self.blacklist_routes = {u'get:/api/v1/ping', u'get:/api/v1/health'}

        # If an app was provided, then call `init_app` for them
        if app is not None:
            self.init_app(app)
        else:
            self.app = None

    def init_app(self, app, config=None):

        # Used passed in config if provided, otherwise use the config from `app`
        if config is not None:
            self.config = config
        elif self.config is None:
            self.config = app.config

        self.app = app

        # Configure any of our middleware
        self.setup_middleware()

    def setup_middleware(self):
        """Helper to configure/setup any Flask-Datadog middleware"""
        # Configure response time middleware (if desired)
        self.app.before_request(self.before_request)
        self.app.after_request(self.after_request)

    def before_request(self):
        """
        middleware handle for before each request
        """
        # Set the request start time
        g.flask_request_start_time = time.time()

    def after_request(self, response):
        """
        middleware handler for after each request
        :param response: the response to be sent to the client
        :type response: ``flask.Response``
        :rtype: ``flask.Response``
        """
        # Return early if we don't have the start time
        if not hasattr(g, 'flask_request_start_time'):
            return response

        if request.url_rule is None:
            return response

        # Get the response time for this request
        elapsed = time.time() - g.flask_request_start_time
        # Convert the elapsed time to milliseconds if they want them
        elapsed = int(round(1000 * elapsed))

        request_method = request.method
        request_path = request.path
        response_status_code = response.status_code

        # encapsulate the vars above, and make a method for reporting stats
        def apply_report():
            try:
                blacklisted = self._is_blacklisted(request_method, request_path)
                if not blacklisted:
                    self.app.logger.debug(
                        "[{}] {} [{}] Time: {}ms".format(request_method, request_path, response_status_code, elapsed))

                    if self.report_method:
                        self.report_method(request_method=request_method, request_path=request_path,
                                           response_status_code=response_status_code, elapsed=elapsed)

            except Exception as e:
                self.app.logger.debug(e, exc_info=True)

        # spawn an async task for reporting our metrics
        self.pool.apply_async(apply_report)

        # We ALWAYS have to return the original response
        return response

    def _is_blacklisted(self, method, url_path):
        return "{}:{}".format(method.lower(), url_path.lower()) in self.blacklist_routes



import json
from functools import wraps

from cerberus import Validator
from flask import request, current_app


def validate_params(schema):
    def decorator(func):
        @wraps(func)
        def wrapper():
            if request.method == 'GET':
                return __handle_get(schema, func)
            else:
                return __handle_other(schema, func)

        return wrapper

    return decorator


def __handle_get(schema, func):
    v = Validator()
    values = dict(request.values.items())
    if not v.validate(values, schema):
        current_app.logger.error({
            'msg': 'Error in validating request-params',
            'errors': v.errors
        })

        return json.dumps({'error': v.errors}), 400

    return func(**v.normalized(values))


def __handle_other(schema, func):
    v = Validator()
    try:
        if not request.is_json:
            error_message = 'Invalid "Content-Type", check your request-headers'
            current_app.logger.error(error_message)

            return json.dumps({'error': error_message}), 400
        else:
            params = dict(request.values.items(), **__get_json())
            if not v.validate(params, schema):
                current_app.logger.error({
                    'msg': 'Error in validating request-params',
                    'errors': v.errors
                })

                return json.dumps({'error': v.errors}), 400
    except Exception:
        error_message = 'Invalid JSON!'
        current_app.logger.error(error_message)

        return json.dumps({'error': error_message}), 400

    return func(**v.normalized(params))


def __get_json():
    if request.data == b"":
        return dict()

    return request.get_json()



import os
import sys
import traceback

import click

from mest.app import Mest

APP_PATH = os.environ.get('MEST_APP')


def require_app_module(ctx):
    error = ctx.obj.get('err')
    if error is not None:
        click.echo('An error occurred while trying to load "{}":'.format(APP_PATH))
        click.echo(error)
        click.echo()
        click.echo(''.join(traceback.format_tb(error.__traceback__)))
        sys.exit(-1)

    app = ctx.obj.get('app')
    if app is None:
        click.echo(f'"{APP_PATH}" could not be found, make sure you run the command from the root directory')
        sys.exit(-1)
    elif app.mest_app is None:
        click.echo(f'could not found mest application instance under ${APP_PATH}')
        sys.exit(-1)

    return app


def require_mest_app(ctx) -> Mest:
    return require_app_module(ctx).mest_app



import logging
import os


class Environment(object):
    TESTING = 'testing'
    DEVELOPMENT = 'development'
    PRODUCTION = 'production'


class MestConfig(object):
    def __init__(self, service_name, listen_port, models_instances=[],
                 local_model_directory_path='_models'):
        self.service_name = service_name
        self.listen_port = listen_port
        self.models_instances = models_instances
        self.local_model_directory_path = local_model_directory_path

        self._setup_general()

    def _setup_general(self):
        self.env = os.environ.get('FLASK_ENV', Environment.DEVELOPMENT)
        self.debug = bool(os.environ.get('DEBUG', False))
        self.log_level = logging.ERROR



import random
import unittest

from mest.models import GenericModel
from tests.common import MyTestModel


class MestCoreModelsTest(unittest.TestCase):
    """Test cases for mest application setup"""

    def setUp(self):
        pass

    def test_simple_predict(self):
        expected_result = 'hello_world'
        test_model = MyTestModel()
        test_model.model = expected_result

        result = test_model.predict()

        self.assertEqual(result, expected_result)

    def test_predict_flow1(self):
        model = ModelWithProcessing1()
        model.init(None)
        expected_result = {
            'prediction': 3 * model.model
        }

        result = model.predict(factor=2)

        self.assertEqual(result, expected_result)

    def test_predict_flow2(self):
        model = ModelWithProcessing2()
        model.init(None)
        expected_result = {
            'prediction': 2 + model.model
        }

        result = model.predict(factor=2)

        self.assertEqual(result, expected_result)


class ModelWithProcessing1(GenericModel):
    def init(self, path):
        self.model = random.randint(1, 10)

    def pre_process(self, factor):
        return self.model * factor

    def infer(self, form_pre_process):
        return self.model + form_pre_process

    def post_process(self, prediction):
        return {
            'prediction': prediction
        }


class ModelWithProcessing2(GenericModel):
    def init(self, path):
        self.model = random.randint(1, 10)

    def pre_process(self, factor):
        return [self.model, factor]

    def infer(self, *form_pre_process):
        return {
            'data': sum(form_pre_process)
        }

    def post_process(self, **prediction):
        return {
            'prediction': prediction['data']
        }



class Status(object):
    SHUTTING_DOWN = 'Shutting Down'
    RUNNING = 'Running'


class __State(object):
    def __init__(self):
        self._shutting_down = False

    def set_to_shutting_down(self):
        self._shutting_down = True

    def is_shutting_down(self):
        return self._shutting_down

    @property
    def status(self):
        return Status.SHUTTING_DOWN if self.is_shutting_down() else Status.RUNNING


runtime_state = __State()



import logging
import sys

from .enrichment import get_request_info
from .formatters import JsonFormatter


def get_stdout_handler(app_name):
    formatter = JsonFormatter(app_name, enrichment_methods=dict(request_info=get_request_info))
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)
    handler.setLevel(logging.WARNING)

    return handler



import json
import random
import unittest

from mest.app import Mest
from mest.app.api import Router, validate_params
from tests.common import generate_mest_config_with_model, MyTestModel
from unittest.mock import patch


class MestCoreAppTest(unittest.TestCase):
    """Test cases for mest application setup"""

    def setUp(self):
        conf = generate_mest_config_with_model()

        self.mest = Mest(conf)

    def test_load_models(self):
        test_model = MyTestModel()
        test_model.model = None

        self.mest.config.models_instances = [test_model]
        self.mest.load_models()

        expected_result = '42'

        self.assertEqual(test_model.model, expected_result)

    def test_add_api(self):
        router = Router('v1')
        expected_result = 'Hello World'

        @router.route(url='ping', method='GET')
        def ping_method():
            return expected_result

        self.mest.register_router(url='/api/v1',
                                  router=router)

        client = self.mest.app.test_client()

        res = client.get('/api/v1/ping')
        self.assertEqual(res.status_code, 200)

        response = res.data.decode('utf-8')
        self.assertEqual(response, expected_result)

    def test_add_api_with_schema(self):
        router = Router('v1')

        @router.route(url='predict')
        @validate_params({'value': {'type': 'integer', 'required': False, 'default': 5}})
        def predict(value, **params):
            return json.dumps({
                'prediction': value
            })

        self.mest.register_router(url='/api/v1',
                                  router=router)

        client = self.mest.app.test_client()

        expected_value = random.sample(population=range(10),
                                       k=1)[0]
        res = client.post('/api/v1/predict',
                          data=json.dumps(dict(value=expected_value)),
                          content_type='application/json')
        self.assertEqual(res.status_code, 200)

        decoded = res.data.decode('utf-8')
        result_dict = json.loads(decoded)
        self.assertEqual(result_dict['prediction'], expected_value)



import click
from ._common import require_mest_app
import sys
import os


class Shell(object):
    @staticmethod
    def register(mest_cli):
        @mest_cli.command('shell', short_help='Run a shell in the app context.')
        @click.pass_context
        def shell(ctx):
            """Run an interactive Python shell in the context of a given
            Flask application.  The application will populate the default
            namespace of this shell according to it's configuration.
            This is useful for executing small snippets of management code
            without having to manually configure the application.
            """

            import code

            mest_app = require_mest_app(ctx)
            mest_app.load_models()
            app = mest_app.app

            banner = 'Python %s on %s\nApp: %s [%s]\nInstance: %s' % (
                sys.version,
                sys.platform,
                app.import_name,
                app.env,
                app.instance_path,
            )
            ctx_d = {}

            # Support the regular Python interpreter startup script if someone
            # is using it.
            startup = os.environ.get('PYTHONSTARTUP')
            if startup and os.path.isfile(startup):
                with open(startup, 'r') as f:
                    eval(compile(f.read(), startup, 'exec'), ctx_d)

            ctx_d.update(app.make_shell_context())

            code.interact(banner=banner, local=ctx_d)



import setuptools

from setuptools import setup

with open('README.md', 'r') as fh:
    long_description = fh.read()

# Pull version from source without importing
# since we can't import something we haven't built yet :)
exec(open('mest/version.py').read())

setup(
    # Needed to silence warnings (and to be pythona worthwhile package)
    name='mest',
    # Needed to actually package something
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=setuptools.find_packages(exclude='tests'),

    # Project URL
    url='https://github.com/orlevi111/mest',

    # author
    author='Or Levi',
    author_email='orlevi128@gmail.com',
    # Needed for dependencies
    install_requires=['Flask==1.1.*', 'cerberus==1.3.*', 'gunicorn==20.*',
                      'python-dotenv==0.10.*', 'click==7.*'],
    # *strongly* suggested for sharing
    version=__version__,
    # The license can be anything you like
    # Choose your license
    license='LICENCE',

    description='A framework for developing a realtime model-inference service.',

    data_files=[('', ['README.md'])],

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Web Environment',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: MIT License',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 3'],

    # We will also need a readme eventually (there will be a warning)
    long_description_content_type='text/markdown',

    long_description=long_description,

    entry_points={"console_scripts": ["mest = mest.cli:main"]}
)



import os
import json
from .dict import DictFormatter


def _is_pretty_log():
    return bool(os.environ.get('PRETTY_LOG', False))


class JsonFormatter(DictFormatter):
    PRETTY_LOG_INDENTION = 4

    def format(self, record):
        message_json = super().format(record)
        indent = self.PRETTY_LOG_INDENTION if _is_pretty_log() else None

        return json.dumps(message_json, indent=indent)



import json
import random
import unittest
from unittest.mock import patch

from mest.app import Mest
from mest.app._state import runtime_state
from mest.app.api import Router
from tests.common import generate_mest_config_with_model, MyTestModel


class MestRouterTest(unittest.TestCase):
    """Test cases for mest pre made api/v1"""

    def setUp(self):
        conf = generate_mest_config_with_model()
        self.model_instance = MyTestModel()

        self.mest_app = Mest(conf)

        self.api_v1 = Router('v1')
        self.api_v1.add_ping_route()
        self.api_v1.add_predict_route(self.model_instance,
                                      schema={'user_id': {'type': 'integer', 'required': True}})
        self.mest_app.register_router(url='/api/v1',
                                      router=self.api_v1)

        self.client = self.mest_app.app.test_client()

    def test_pre_made_api_ping(self):
        res = self.client.get('/api/v1/ping')
        self.assertEqual(res.status_code, 200)

        decoded = res.data.decode('utf-8')
        result_dict = json.loads(decoded)
        expected_text = 'Pong from {}!'.format(self.mest_app.config.service_name)
        self.assertIn(expected_text, result_dict['message'])

    def test_pre_made_api_ping_when_shutting_down(self):
        runtime_state._shutting_down = True
        res = self.client.get('/api/v1/ping')
        self.assertEqual(res.status_code, 503)

        decoded = res.data.decode('utf-8')
        result_dict = json.loads(decoded)
        expected_text = 'Pong from {}!'.format(self.mest_app.config.service_name)
        self.assertIn(expected_text, result_dict['message'])
        runtime_state._shutting_down = False

    @patch('tests.common.MyTestModel.predict')
    def test_add_predict(self, mock_predict):
        client = self.mest_app.app.test_client()

        expected_result = 'hello world'
        mock_predict.return_value = expected_result

        random_user_id = random.sample(population=range(10),
                                       k=1)[0]
        res = client.post('/api/v1/predict',
                          data=json.dumps(dict(user_id=random_user_id)),
                          content_type='application/json')

        decoded = res.data.decode('utf-8')
        result_value = json.loads(decoded)

        self.assertEqual(res.status_code, 200)
        self.assertEqual(result_value, expected_result)
        self.assertEqual(mock_predict.call_count, 1)



import os
import sys

import click


class Test(object):
    @staticmethod
    def register(mest_cli):
        @mest_cli.command(name='test', help='Run unit tests')
        def test():
            if not os.path.isdir('./tests'):
                click.echo('Could not find "tests" package,' +
                           ' please make sure "tests" exists and is importable (contains __init__.py)')
                sys.exit(1)

            os.environ.setdefault('FLASK_ENV', 'testing')

            import unittest
            loader = unittest.TestLoader()
            tests = loader.discover('./tests', pattern="*.py")
            test_runner = unittest.runner.TextTestRunner(verbosity=2)
            result = test_runner.run(tests)

            has_errors = any(result.errors + result.failures)
            code = -1 if has_errors else 0

            return sys.exit(code)



import json
import logging
import os
import signal

from flask import Flask

from ._logs.initializers import get_stdout_handler
from ._middlewares.response_time import RequestStatter
from ._state import runtime_state
from .api import Router
from .mest_conf import MestConfig, Environment

os.environ.setdefault('MEST_APP', 'app.py')


class Mest(object):
    def __init__(self, config: MestConfig):
        self.config = config
        self.app: Flask = self.__create_flask_app()
        self.__init_logger()
        self._setup = False

    def setup(self):
        if self.__is_cli:
            self._expose_app_for_gunicorn()
            return self
        if self._setup:
            return self

        _handle_sigterm(self.logger)
        _register_error_handler(self.app)

        self.load_models()

        self._setup = True

        return self

    def run(self):
        if not self._setup:
            raise RuntimeError('setup() must be called before run()')

        self.app.run(port=self.config.listen_port,
                     host='0.0.0.0')

    def register_router(self, url: str, router: Router):
        self.app.register_blueprint(router._blueprint, url_prefix=url)

    @property
    def logger(self):
        return self.app.logger

    def __create_flask_app(self) -> Flask:
        app = Flask(__name__)
        app.config['MEST'] = self.config

        # stats middleware
        RequestStatter(app, service_name=self.config.service_name)

        # logger
        if not self.config.debug:
            app.logger.handlers = []

        stdout_handler = get_stdout_handler(self.config.service_name)
        if self.config.env == Environment.DEVELOPMENT:
            stdout_handler.setLevel(logging.DEBUG)

        if not self.config.debug:
            app.logger.addHandler(stdout_handler)

        app.logger.setLevel(logging.DEBUG)

        return app

    def load_models(self):
        model_instance_name = None
        try:
            for model_instance in self.config.models_instances:
                model_instance_name = type(model_instance).__name__
                model_instance.init(self.config.local_model_directory_path)
                self.logger.debug('Initialized model {} successfully'.format(model_instance_name))
        except Exception:
            self.app.logger.exception('Could not initialize model: {}'.format(model_instance_name))
            if self.config.env == Environment.PRODUCTION:
                raise

    @property
    def __is_cli(self) -> bool:
        cli = os.environ.get('MEST_CLI', '')

        return cli.lower() == 'true'

    def __init_logger(self):
        logger = logging.getLogger(self.config.service_name)
        logger.setLevel(self.logger.level)
        logger.handlers = self.logger.handlers

    def _expose_app_for_gunicorn(self):
        main_app_module = os.environ.get('MEST_APP').split('.')[0]
        main_module = __import__(main_app_module)
        main_module.app = self.app


def _handle_sigterm(logger):
    def sigterm_handler(signum, frame):
        logger.info('SIGTERM ACCEPTED')
        runtime_state.set_to_shutting_down()

    signal.signal(signal.SIGTERM, sigterm_handler)


def _register_error_handler(app):
    @app.errorhandler(Exception)
    def generic_exception_handler(error):
        app.logger.exception(error)

        return json.dumps({'message': str(error)}), 500

    @app.errorhandler(404)
    def not_found(error):
        app.logger.info(error)

        return json.dumps({'message': '{}'.format(error)}), 404
